package edu.wlu.cs.levy.RAAM;

import java.util.Vector;
import java.util.Random;
import java.util.Hashtable;

import edu.wlu.cs.levy.SNARLI.BPLayer;


/**
 * IFSDecoder is a class supporting multi-dimensional IFS
 * attractor and tree computations by Monte-Carlo sampling. Based on
 *
 * <PRE>&#064PhdThesis{LevyThesis,
 *  author =          {S.D. Levy},
 *  title =           {Infinite RAAM: Initial Explorations into a Fractal 
 *                     Basis for Cognition},
 *  school =          {Brandeis University},
 *  year =            {2002},
 *  address =         {Waltham, Mass.},
 * }
 *  </PRE>
 *
 * @author      Simon Levy
 * @version     %I%, %G%
 * @since JDK1.2 
 */
public class IFSDecoder {

    // transforms count
    private int _ntrans;

    // dimensions count
    private int _ndims;

    // input, output layers
    private BPLayer _inp, _out;
    

    /**
     * Creates an IFSDecoder using weights and biases. (Simon's
     * preference) Weights array should have D rows and K*D columns,
     * where D is the number of dimensions and K the number of
     * transforms.  Bias vector should have K*D elements.
     *
     * @param w weights
     * @param b biases
     * @param r random-number generator
     *
     */
    public IFSDecoder(double [][] w, double [] b) {

        int d = w[0].length; 
        int k = b.length / d;

        _inp = new BPLayer(d);
        _out = new BPLayer(k*d);
        _out.connect(_inp);

        // system params
        _ntrans = k;
        _ndims = d;

        _out.setWeights(_inp, w);
        _out.setBias(b);
    }

    /**
     * Creates an IFSDecoder using one-dimensional array of
     * weights. (Anthony's preference) Weights are linearized by read
     * Pollack-style IFS weights one transform at a time.
     *
     * @param w linearized IFSDecoder weights
     * @param k number of IFSDecoder transforms
     *
     */
    public IFSDecoder(double [] w, int k) {
        this(linear2ifs(w, k));
    }
    



    /**
     * Creates an IFSDecoder, using K x (D^2+D) array of IFS transform
     * weights, where K is # of transforms, D is # of
     * dimensions. (Pollack's preference) Weights are linearized by
     * taking Pollack-style rectangular weights array and reading top
     * to bottom, left to right.
     *
     * @param w IFS transform weights
     * @param r random-number generator
     * 
     */
    public IFSDecoder(double [][] w) {

        this(ifs2weights(w), ifs2bias(w));
    }

    /**
     * Construct decoder from RAAM network.
     * @param raam RAAM network
     */
    public IFSDecoder(RAAM raam) {
        this(raam.out.getWeights(raam.hid), raam.out.getBias());
    }


    /**
     * Returns attractor points as array.  Uses Monte Carlo sampling
     * of attractor approximations, starting with the unit hypercube.
     * In general this isn't of much use for more than three
     * dimensions, because of the difficulty of visualizing
     * high-dimensional objects.
     *
     * @param m number of sample points
     * @param n number of IFS iterations
     * @param r random-number generator
     * @return MxD array of M D-dimensional attractor points
     */
    public double [][] getAttractor(int m, int n, Random r) {

        return getattr(m, n, r);
    }

    /**
     * Returns attractor points as array.  Uses Monte Carlo sampling
     * of attractor approximations, starting with the unit hypercube.
     * In general this isn't of much use for more than three
     * dimensions, because of the difficulty of visualizing
     * high-dimensional objects.
     *
     * @param m number of sample points
     * @param n number of IFS iterations
     * @param seed seed for random-number generator
     * @return MxD array of M D-dimensional attractor points
     */
    public double [][] getAttractor(int m, int n, long seed) {

        return getAttractor(m, n, new Random(seed));
    }

    /**
     * Returns attractor points as array.  Uses Monte Carlo sampling
     * of attractor approximations, starting with the unit hypercube.
     * In general this isn't of much use for more than three
     * dimensions, because of the difficulty of visualizing
     * high-dimensional objects.
     *
     * @param m number of sample points
     * @param n number of IFS iterations
     * @return MxD array of M D-dimensional attractor points
     */
    public double [][] getAttractor(int m, int n) {

        return getAttractor(m, n, new Random());
    }


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param d dictionary for labeling points
     * @param e epsilon distance from dictionary entries
     * @param r random-number generator
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, Vectionary d, double e, Random r) 
        throws RuntimeException {
	check_dims(d);
        return get_trees(m, n, d, e, r);
    }        


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param d dictionary for labeling points
     * @param e epsilon distance from dictionary entries
     * @param seed seed for random-number generator
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, Vectionary d, double e, long seed) 
        throws RuntimeException {

        return getTrees(m, n, d, e, new Random(seed));
    }        


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param d dictionary for labeling points
     * @param e epsilon distance from dictionary entries
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, Vectionary d, double e) 
        throws RuntimeException {

        return getTrees(m, n, d, e, new Random());
    }        


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D. Trees
     * terminate in default terminal 'X'.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param e epsilon distance from attractor
     * @param r random-number generator
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, double e, Random r) 
        throws RuntimeException {

        return get_trees(m, n, new Vectionary(_ndims), e, r);
    }        


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D. Trees
     * terminate in default terminal 'X'.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param d dictionary for labeling points
     * @param e epsilon distance from dictionary entries
     * @param seed seed for random-number generator
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, double e, long seed) 
        throws RuntimeException {

        return getTrees(m, n, e, new Random(seed));
    }        


    /**
     * Returns array of M trees by Monte Carlo sampling in [0,N)^D. Trees
     * terminate in default terminal 'X'.
     * @param m number of points to sample
     * @param n number of IFS iterations (= max tree depth)
     * @param e epsilon distance from attractor
     * @return array of unique trees
     */
    public Tree [] getTrees(int m, int n, double e) 
        throws RuntimeException {

        return getTrees(m, n, e, new Random());
    }        

    // dimensionality check
    private void check_dims(Vectionary d) throws RuntimeException {
        if (d.width() != _ndims) {
            throw new RuntimeException("Dictionary entry size " + d.width() +
                                       " mismatches dimension size " + _ndims);
        }
    }


    // common method for all getTrees()
    private Tree [] get_trees(int m, int n, Vectionary d, double e, Random r) {
        //double [][] a = getattr(m, n, r);
	Vector v = new Vector();
        for (int i=0; i<m; ++i) {
            double [] x = randvec(_ndims, r);
	    Tree t = decode(x, d, /*a,*/ 0, n, e);
	    if (t != null) v.add(t);
        }
        Tree [] trees = new Tree[v.size()];
        for (int i=0; i<v.size(); ++i) {
	    trees[i] = (Tree)v.elementAt(i);
        }
        return trees;
    }

    // decodes point to tree
    private Tree decode(double [] x, Vectionary dict, /*double [][] attr, */
                        int depth, int maxdepth, double epsilon)
        throws RuntimeException {

        // terminate on attractor or at maximum depth
        if (depth == maxdepth /* || onattr(x, attr, epsilon)*/) {
            String s = dict.closest(x, epsilon);
            return s == null ? null : new Tree(s);
        }

        // recur off attractor
        else {
            Tree [] branches = new Tree[_ntrans];
            for (int k=0; k<_ntrans; ++k) {
                double [] tx = transform(k, x);
		Tree branch = decode(tx, dict, /*attr,*/ depth+1, maxdepth, 
				     epsilon);
		if (branch == null) return null;
                branches[k] = branch;
                                     
            }
            return new Tree(branches);
        }
    }

    // returns true if point is on attractor, false otherwise
    private boolean onattr(double [] x, double [][] a, double epsilon) {
        for (int i=0; i<a.length; ++i) {
            if (eucdist(x, a[i]) < epsilon) return true;
        }
        return false;
    }

    // returns Kth transform of point X
    private double [] transform(int k, double [] x) {
        double [] y = _out.activate(_inp, x);
        double [] tx = new double [_ndims];
        for (int d=0; d<_ndims; ++d) {
            tx[d] = y[k*_ndims+d]; 
        }
        return tx;
    }

    // returns attractor as an array of vectors
    private double [][] getattr(int m, int n, Random r) {

        // initial attractor : M random points in unit hypercube
        double [][] a = randmat(m, _ndims, r);

        // run IFS N times
        for (int i=0; i<n; ++i) {
            a = condense(a, r);
        }

        return a;

    }

    // runs one IFS iteration, returning union of transforms over input
    private double [][] condense(double [][] a, Random r) {

        int m = a.length;

        double [][] b = new double [m][];

        for (int i=0; i<m; ++i) {
            double [] x = a[i];
            int k = r.nextInt(_ntrans); // pick a random transform
            b[i] = transform(k, x);
        }

        return b;
    }

    // returns Euclidean distance between two vectors
    private static double eucdist(double [] x, double [] y) {
        double dist = 0;
        for (int j=0; j<x.length; ++j) {
            double diff = x[j] - y[j];
            dist += diff*diff;
        }
        return Math.sqrt(dist);
    }

    // return a random vector in [0,1)
    private static double [] randvec(int n, Random r) {
        double [] x = new double [n];
        for (int j=0; j<n; ++j) {
            x[j] = r.nextDouble();
        }
        return x;
    }

    // return a random matrix in [0,1)
    private static double [][] randmat(int m, int n, Random r) {
        double [][] a = new double [m][];
        for (int i=0; i<m; ++i) {
            a[i] = randvec(n, r);
        }
        return a;
    }

    // returns a terminal tree with arbitrary symbol
    private static Tree terminal_tree() {
        return new Tree("X");
    }

    // extract IFS transform weights from one-dimensional array
    private static double [][] linear2ifs(double [] w, int K) {
        double [][] ww = new double [K][w.length/K];
        int i = 0;
        for (int k=0; k<K; ++k) {
            for (int j=0; j<w.length/K; ++j) {
                ww[k][j] = w[i++];
            }
        }
        return ww;
    }

    // extract weights from IFS transforms array
    private static double [][] ifs2weights(double [][] w) {
        int K = w.length, D = length2dims(w[0].length);
        double [][] ww = new double [K*D][D];
        for (int i=0; i<K; ++i) {
            for (int d1=0; d1<D; ++d1) {
                for (int d2=0; d2<D; ++d2) {
                    ww[i*D+d1][d2] = w[i][d1*(D+1)+d2];
                }
            }
        }
        return ww;
    }

    // extract biases from IFS transforms array
    private static double [] ifs2bias(double [][] w) {
        int K = w.length, D = length2dims(w[0].length);
        double [] b = new double [K*D];
        int i = 0;
        for (int k=0; k<K; ++k) {
            for (int d=0; d<D; ++d) {
                b[i++] = w[k][d*(D+1)+D];
            }
        }
        return b;
    }

    // compute # of dimensions based on # of coefficients in one transform
    private static int length2dims(int len) {
        int d = 1;
        for (; d*(d+1)<len; d++)
            ;
        return d;
    }
}
