package edu.wlu.cs.levy.RAAM;

import java.util.Vector;
import java.util.Random;

import edu.wlu.cs.levy.SNARLI.BPLayer;
import edu.wlu.cs.levy.SNARLI.SNVector;

/** 
 * Pollack's Recursive Auto-Associative Memory using SNARLI.  Based on
 *   <PRE>&#064Article{Pollack90,
 *   author = 	 {Pollack, J.B.},
 *   title = 	 {Recursive distributed representations},
 *   journal = 	 {Artifical Intelligence},
 *   year = 	 {1990},
 *   volume = 	 {36},
 *   pages = 	 {77-105}
 * }
 *  </PRE>
 *
 * This version of RAAM differs from Pollack's in using a terminal test based
 * on proximity to dictionary entries, rather than an absolute threshold.
 */

public class RAAM {
  
    private final static int MAXDEPTH = 50;  // maximum recursion depth
    private final static double TOL = 0.1;   // tolerance for terminal test

    protected BPLayer hid, out;              // accesses by IFSDecoder

    protected BPLayer inp;
    private int nhid;
    private int arity;

    /**
     * Creates a RAAM encoder/decoder network with specified number of 
     * hidden units.  Number of inputs = number of outputs = twice this
     * number.  Initializes network weights and biases to small random
     * values.
     * @param nhid number of hidden units
     * @param arity fixed tree arity
     */
    public RAAM(int nhid, int arity){
	
	this(nhid, arity, new Random());
    }

    /**
     * Creates a RAAM encoder/decoder network with specified number of 
     * hidden units.  Number of inputs = number of outputs = twice this
     * number.  Initializes network weights and biases to small random
     * values determined by random-number generator seed.
     * @param nhid number of hidden units
     * @param arity fixed tree arity
     * @param seed seed for random-number generator
     */
    public RAAM(int nhid, int arity, long seed){
	
	this(nhid, arity, new Random(seed));
    }

    /**
     * Creates a RAAM encoder/decoder network with specified number of 
     * hidden units.  Number of inputs = number of outputs = twice this
     * number.  Initializes network weights and biases to small random
     * values determined by random-number generator.
     * @param nhid number of hidden units
     * @param arity fixed tree arity
     * @param rand random-number generator
     */
    public RAAM(int nhid, int arity, Random rand) {

	this.nhid = nhid;
	this.arity = arity;

	inp = new BPLayer(nhid*arity);
	hid = new BPLayer(nhid);
	out = new BPLayer(nhid*arity);

	hid.connect(inp);
	out.connect(hid);

	hid.randomize(rand);
	out.randomize(rand);
    }
    
    /**
     * Trains the RAAM network on a set of trees.
     * @param trees trees to learn
     * @param dict dictionary mapping from terminals to fixed-size arrays
     * @param nep number of epochs
     * @param eta learning rate
     * @param mu momentum
     * @param report epochs between error reports
     * @throws IllegalArgumentException if trees aren't binary 
     * @throws IllegalArgumentException if size of any dictionary entry
     * mismatches number of hidden units
     */
    public void train(Tree [] trees, Vectionary dict, int nep, 
		      double eta, double mu, int report) 

	throws IllegalArgumentException {

	// reset previous weight,bias changes for momentum
	hid.resetMu();
	out.resetMu();

	for(int i=0; i<nep; ++i) {
	
	    // training patterns change on every iteration
	    double [][] pats = _get_patterns(trees, dict);

	    inp.attach(pats);
	    out.attach(pats);

	    double rmserr = out.getRMSError();

	    out.batch(eta, mu);
	    
	    // report rms error every 1000 iterations
	    BPLayer.reportValue(i, nep, report, rmserr, System.err);
	}
    }

    /**
     * Encodes a tree to a fixed-size vector.
     * @param tree tree to encode
     * @param dict dictionary mapping from terminals to fixed-size arrays
     * @return fixed-size vector encoding tree
     * @throws IllegalArgumentException if tree isn't binary 
     * @throws IllegalArgumentException if size of any dictionary entry
     * mismatches number of hidden units
     */
    public double [] encode(Tree tree, Vectionary dict) 

	throws IllegalArgumentException {

	if(tree.isTerminal()){
	    throw new IllegalArgumentException("Tree is terminal: " + tree);
	}

	return _encode(tree, dict);
    }


    /**
     * Decodes a fixed-size vector to a tree.
     * @param code fixed-size vector encoding
     * @param dict dictionary mapping from terminals to fixed-size arrays
     * @return decoded tree
     * @throws IllegalArgumentException if size of code or dictionary entry
     * mismatches number of hidden units
     */
    public Tree decode(double [] code, Vectionary dict) 

	throws IllegalArgumentException {
	
	return _decode(code, dict, 0);
    }

    /**
     * Returns decoder weights, biases as IFS (one row per transform)
     * @return IFS weights
     */
    public double [][] getDecoder() {

	double [][] ww = new double [arity][nhid*(nhid+1)];

	double [][] w = out.getWeights(hid);

	double [] b = out.getBias();

	int i = 0;
	for (int k=0; k<arity; ++k) {
	    for (int d1=0; d1<nhid; ++d1) {
		for (int d2=0; d2<nhid; ++d2) {
		    ww[k][d1*(nhid+1)+d2] = w[d1][k*nhid+d2];
		}
		ww[k][d1*(nhid+1)+nhid]= b[i++];
	    }
	}

	return ww;
    }

    // decode tree with depth checking
    private Tree _decode(double [] code, Vectionary dict, int depth) 

	throws IllegalArgumentException {

	Tree t = null;

	if(depth > MAXDEPTH) {
	    throw new 
	     IllegalArgumentException("exceeded maximum depth of " + MAXDEPTH);
	}
	
	if(_is_terminal(code, dict)) {
	    String s = dict.closest(code, Double.POSITIVE_INFINITY);
	    t = new Tree(s);
	}

	else {
	    try {

	      double [] actout = out.activate(hid, code);

	      Tree [] branches = new Tree [arity];

	      for(int k=0; k<arity; ++k) {
		  double [] subcode = _split(actout, k*nhid, (k+1)*nhid);
		  branches[k] = _decode(subcode, dict, depth+1);
		  if(branches[k] == null) return null;
	      }

	      t = new Tree(branches); 

	    }
	    catch(Exception e) {
		System.err.println(e);
	    }
	}

	return t;
    }

    // mutually recursive with above
    private double [] _encode(Tree tree, Vectionary dict) 

	throws IllegalArgumentException {

	double [] input = new double [nhid*arity];

	for(int k=0; k<arity; ++k) {
	    double [] code = _get_code(tree.getBranch(k), dict);
	    _concat(input, code, k);
	}
        
	return hid.activate(inp, input);
    }

    // returns an array of patterns for the encoding of each node of each tree
    private double [][] _get_patterns(Tree [] trees, Vectionary dict)
	throws IllegalArgumentException {

	Vector v = new Vector();
	for(int i=0; i<trees.length; ++i) {
	  _sub_patterns(trees[i], v, dict);
	}
	double [][] pats = new double [v.size()][nhid*arity];
	for(int i = 0; i < v.size(); i++){
	    pats[i] =(double [])v.elementAt(i);
	}

	return pats;
    }

    // recursive portion of the _get_patterns method
    private void _sub_patterns(Tree t, Vector v, Vectionary dict)
	throws IllegalArgumentException {

	if(!t.isTerminal()){

	    double [] input = new double [nhid*arity];

	    for(int k=0; k<arity; ++k) {
		double [] code =  _get_code(t.getBranch(k), dict);
		_concat(input, code, k);
	    }

	    v.add(input);

	    for(int k=0; k<arity; ++k) {
		_sub_patterns(t.getBranch(k), v, dict);
	    }
	}
    }

    // returns the encoding for a Tree
    private double [] _get_code(Tree tree, Vectionary dict)
	throws IllegalArgumentException {

	return tree.isTerminal() ? 
	    _tcode(tree.getValue(), dict) :
	    _encode(tree, dict);
    }

    // returns terminal code for string
    private static double [] _tcode(String s, Vectionary dict) { 
	double [][] codes = dict.lookup(s);
	return codes[0]; // RAAM only uses one code per terminal
    }

    // terminal test based on proximity to dictionary entry
    private static boolean _is_terminal(double [] code, Vectionary dict) {
	String s = dict.closest(code, Double.POSITIVE_INFINITY);
	double [] closest = _tcode(s, dict);
	return(new SNVector(closest)).eucDist(new SNVector(code)) < TOL;
    }

    // make new vector by concatenating
    private static void _concat(double [] a, double [] b, int k) 
	throws RuntimeException {

	for (int i=0; i<b.length; ++i) {
	    int j = k*b.length + i;
	    if (j == a.length) {
		throw new RuntimeException("Not enough hidden units");
	    }
	    a[j] = b[i];
	}
    }

    // split vector in specified bounds
    private static double [] _split(double [] a, int beg, int end) {
	double [] b = new double [end-beg];
	for(int i=beg; i<end; ++i) b[i-beg] = a[i];
	return b;
    }

}
